@@ -20,7 +20,6 @@
 
 from __future__ import absolute_import
 
-import collections
 import datetime
 import gflags
 import inspect
@@ -29,6 +28,7 @@ import time
 import threading
 import logging
 
+from . import backend
 from . import common_defs
 from . import kbevent
 from . import kegnet
@@ -41,12 +41,6 @@ from kegbot.util import util
 
 FLAGS = gflags.FLAGS
 
-class TapManagerError(Exception):
-  """ Generic TapManager error """
-
-class UnknownTapError(TapManagerError):
-  """ Raised when tap requested does not exist """
-
 
 def EventHandler(event_type):
   def decorate(f):
@@ -88,22 +82,14 @@ class TapManager(Manager):
   taps.
   """
 
-  def __init__(self, event_hub, backend):
+  def __init__(self, event_hub, backend_obj):
     super(TapManager, self).__init__(event_hub)
-    self._backend = backend
+    self._backend = backend_obj
     self._taps = {}
-    self._meters = {}
-
-  def TapExists(self, name):
-    return name in self._taps
 
   def GetAllTaps(self):
     return self._taps.values()
 
-  def _CheckTapExists(self, name):
-    if not self.TapExists(name):
-      raise UnknownTapError
-
   def _RegisterOrUpdateTap(self, name, ml_per_tick, relay_name=None):
     existing = self._taps.get(name)
     new_tap = Tap(name, ml_per_tick, relay_name)
@@ -111,58 +97,22 @@ class TapManager(Manager):
       return
     self._logger.info('Updating tap: %s' % new_tap)
     self._taps[name] = new_tap
-    max_delta = (1 / ml_per_tick) * 500.0 # TODO(mikey): remove
-    self._meters[name] = FlowMeter(name, max_delta)
 
   def _RemoveTap(self, name):
+    tap = self._taps.get(name)
+    self.logger.info('Removing tap: %s' % tap)
     del self._taps[name]
     del self._meters[name]
 
   def GetTap(self, name):
-    """Returns the registered tap identified by `name`."""
-    self._CheckTapExists(name)
-    return self._taps[name]
+    """Returns the registered tap identified by `name`, or None."""
+    return self._taps.get(name)
 
-  def GetMeter(self, name):
-    """Returns the meter for the registered tap identified by `name`."""
-    self._CheckTapExists(name)
-    return self._meters[name]
+  @EventHandler(kbevent.SyncEvent)
+  def _HandleSync(self, event):
+    new_taps = event.data.get('taps', [])
 
-  def UpdateDeviceReading(self, tap_name, value, when=None):
-    """Updates the meter reading at the tap named `tap_name`.
-
-    Args
-      tap_name: the tap name
-      value: the instantaneous meter reading
-      when: the timestamp of the meter reading, or None for now
-
-    Returns
-      the amount the meter was incremented (as in FlowMeter.SetTicks)
-    """
-    meter = self.GetMeter(tap_name)
-    delta = meter.SetTicks(value)
-    return delta
-
-  @EventHandler(kbevent.StartedEvent)
-  @EventHandler(kbevent.HeartbeatMinuteEvent)
-  def _HandleHeartbeat(self, event):
-    taps = None
-
-    try:
-      taps = self._backend.GetAllTaps()
-    except requests.exceptions.ConnectionError, e:
-      self._logger.error('Kegbot backend was unreachable due to socket error: '
-          '%s' % e)
-    except kbapi.ServerError, e:
-      self._logger.error('Kegbot API backend returned a server error: %s' % e)
-
-    if taps is None:
-      self._logger.error('Could not sync taps. Is --api_url correct? (current=%s)' % FLAGS.api_url)
-      return
-
-    for tap in taps:
-      if self.TapExists(tap.meter_name):
-        continue
+    for tap in new_taps:
       self._RegisterOrUpdateTap(tap.meter_name, tap.ml_per_tick,
           relay_name=tap.relay_name)
 
@@ -171,6 +121,7 @@ class FlowManager(Manager):
   def __init__(self, event_hub, tap_manager):
     super(FlowManager, self).__init__(event_hub)
     self._tap_manager = tap_manager
+    self._meters = {}
     self._flow_map = {}
     self._logger = logging.getLogger("flowmanager")
     self._next_flow_id = int(time.time())
@@ -186,6 +137,14 @@ class FlowManager(Manager):
     self._next_flow_id += 1
     return ret
 
+  @util.synchronized
+  def GetMeter(self, meter_name):
+    m = self._meters.get(meter_name)
+    if not m:
+      m = FlowMeter(meter_name, 1000.0)
+      self._meters[meter_name] = m
+    return m
+
   def GetActiveFlows(self):
     return self._flow_map.values()
 
@@ -197,23 +156,18 @@ class FlowManager(Manager):
   def GetFlow(self, tap_name):
     return self._flow_map.get(tap_name)
 
-  def StartFlow(self, tap_name, username='', max_idle_secs=10):
-    """Starts a new flow on the given tap, or takes over the existing flow.
+  def StartFlow(self, meter_name, username='', max_idle_secs=10):
+    """Starts a new flow on the given meter, or takes over the existing flow.
 
     Args
-      tap_name: name of the tap
+      meter_name: name of the meter producing the flow
       username: username to own the flow, or None/empty string for anonymous
       max_idle_secs: maximum number of seconds until the flow is marked idle
 
     Returns
       Tuple of (Flow, boolean is_new).
-
-    Raises
-      UnknownTapError: if the named tap does not exist
     """
-    tap = self._tap_manager.GetTap(tap_name)
-
-    current = self.GetFlow(tap_name)
+    current = self.GetFlow(meter_name)
     if current and username:
       if current.GetUsername() == username:
         # Already have a flow for this user; no change.
@@ -228,12 +182,12 @@ class FlowManager(Manager):
       else:
         self._logger.info('User "%s" is replacing the existing flow' %
             username)
-        self.StopFlow(tap_name)
+        self.StopFlow(meter_name)
 
     # Start a new flow.
-    new_flow = Flow(tap, flow_id=self._GetNextFlowId(), username=username,
+    new_flow = Flow(meter_name, flow_id=self._GetNextFlowId(), username=username,
         max_idle_secs=max_idle_secs)
-    self._flow_map[tap_name] = new_flow
+    self._flow_map[meter_name] = new_flow
     self._logger.info('Starting flow: %s' % new_flow)
     self._PublishUpdate(new_flow)
 
@@ -242,33 +196,28 @@ class FlowManager(Manager):
       self._PublishRelayEvent(new_flow, enable=True)
     return new_flow, True
 
-  def StopFlow(self, tap_name):
-    """Ends the flow at the given tap name.
+  def StopFlow(self, meter_name):
+    """Ends the flow at the given meter name.
 
     Returns
       the previously active flow, or None if no flow was active.
-
-    Raises
-      UnknownTapError: if the named tap doesn't exist
     """
-    try:
-      flow = self.GetFlow(tap_name)
-    except UnknownTapError:
-      return None
+    flow = self.GetFlow(meter_name)
     if not flow:
+      self._logger.warning('No flow to stop on meter %s' % meter_name)
       return None
 
     self._logger.info('Stopping flow: %s' % flow)
     self._PublishRelayEvent(flow, enable=False)
-    del self._flow_map[tap_name]
+    del self._flow_map[meter_name]
     self._StateChange(flow, kbevent.FlowUpdate.FlowState.COMPLETED)
     return flow
 
-  def UpdateFlow(self, tap_name, meter_reading, when=None):
-    """Creates or updates a flow at `tap_name`.
+  def UpdateFlow(self, meter_name, meter_reading, when=None):
+    """Creates or updates a flow at `meter_name`.
 
     Args
-      tap_name: name of the tap to update
+      meter_name: name of the tap to update
       meter_reading: instantaneous meter reading
       when: timestamp used for activity (defaults to datetime.now)
 
@@ -276,22 +225,16 @@ class FlowManager(Manager):
       Tuple of (flow, is_new).
       Flow may be None if no update occurred.
     """
-    try:
-      tap = self._tap_manager.GetTap(tap_name)
-    except TapManagerError:
-      # tap is unknown or not available
-      # TODO(mikey): guard against this happening
-      return None, None
-
-    delta = self._tap_manager.UpdateDeviceReading(tap.GetName(), meter_reading)
+    meter = self.GetMeter(meter_name)
+    delta = meter.SetTicks(meter_reading)
     self._logger.debug('Flow update: tap=%s meter_reading=%i (delta=%i)' %
-        (tap_name, meter_reading, delta))
+        (meter_name, meter_reading, delta))
 
     is_new = False
-    flow = self.GetFlow(tap_name)
+    flow = self.GetFlow(meter_name)
     if flow is None:
       self._logger.debug('Starting flow implicitly due to activity.')
-      flow, is_new = self.StartFlow(tap_name)
+      flow, is_new = self.StartFlow(meter_name)
     if when is None:
       when = datetime.datetime.now()
 
@@ -309,28 +252,28 @@ class FlowManager(Manager):
 
   @EventHandler(kbevent.MeterUpdate)
   def HandleFlowActivityEvent(self, event):
-    try:
-      flow_instance, is_new = self.UpdateFlow(event.tap_name, event.reading)
-    except UnknownTapError:
-      return None
+    flow_instance, is_new = self.UpdateFlow(event.meter_name, event.reading)
 
   @EventHandler(kbevent.HeartbeatSecondEvent)
   def _HandleHeartbeatEvent(self, event):
     for flow in self.GetActiveFlows():
-      tap = flow.GetTap()
       if flow.IsIdle():
         self._logger.info('Flow has become too idle, ending: %s' % flow)
         self._StateChange(flow, kbevent.FlowUpdate.FlowState.IDLE)
-        self.StopFlow(flow.GetTap().GetName())
+        self.StopFlow(flow.GetMeterName())
       else:
-        if tap.GetRelayName():
-          if flow.GetUsername():
-            self._PublishRelayEvent(flow, enable=True)
+        if flow.GetUsername():
+          self._PublishRelayEvent(flow, enable=True)
 
   def _PublishRelayEvent(self, flow, enable=True):
     self._logger.debug('Publishing relay event: flow=%s, enable=%s' % (flow,
         enable))
-    tap = flow.GetTap()
+    tap = self._tap_manager.GetTap(flow.GetMeterName())
+    if not tap:
+      # Unknown meter; don't attempt to enable any relays for it
+      # since we don't know its configuration.
+      return
+
     relay = tap.GetRelayName()
     if not relay:
       self._logger.debug('No relay for this tap')
@@ -347,16 +290,16 @@ class FlowManager(Manager):
   @EventHandler(kbevent.FlowRequest)
   def _HandleFlowRequestEvent(self, event):
     if event.request == event.Action.START_FLOW:
-      self.StartFlow(event.tap_name)
+      self.StartFlow(event.meter_name)
     elif event.request == event.Action.STOP_FLOW:
-      self.StopFlow(event.tap_name)
+      self.StopFlow(event.meter_name)
 
 
 class DrinkManager(Manager):
-  def __init__(self, event_hub, backend):
+  def __init__(self, event_hub, backend_obj):
     super(DrinkManager, self).__init__(event_hub)
-    self._backend = backend
-    self._pending = collections.deque()
+    self._backend = backend_obj
+    self._pending = []
 
   @EventHandler(kbevent.FlowUpdate)
   def HandleFlowUpdateEvent(self, event):
@@ -371,22 +314,28 @@ class DrinkManager(Manager):
     self._FlushPending()
 
   def _FlushPending(self):
-    while len(self._pending):
-      event = self._pending.popleft()
+    if not self._pending:
+      return
+
+    pending = self._pending[:]
+    del self._pending[:]
+
+    self._logger.info('Posting %s pending event(s)' % len(pending))
+
+    for event in pending:
       try:
         self._PostDrink(event)
-      except Exception, e:
+      except backend.BackendException as e:
         self._logger.warning('Error posting drink: %s' % e)
-        self._pending.appendleft(event)
-        break
+        self._pending.append(event)
 
   def _PostDrink(self, event):
-    self._logger.info('Processing pending drink: flow_id=0x%08x' % event.flow_id)
+    self._logger.info('Processing pending drink: flow_id=0x%08x, meter=%s' % (
+      event.flow_id, event.meter_name))
 
     ticks = event.ticks
     username = event.username
-    volume_ml = event.volume_ml
-    tap_name = event.tap_name
+    meter_name = event.meter_name
     pour_time = event.last_activity_time
     duration = (event.last_activity_time - event.start_time).seconds
     flow_id = event.flow_id
@@ -394,17 +343,20 @@ class DrinkManager(Manager):
     # TODO: add to flow event
     auth_token = None
 
-    if volume_ml <= common_defs.MIN_VOLUME_TO_RECORD:
-        self._logger.info('Not recording flow: volume (%i mL) <= '
-            'MIN_VOLUME_TO_RECORD (%i)' % (volume_ml, common_defs.MIN_VOLUME_TO_RECORD))
+    if ticks <= 0:
+        self._logger.info('Not recording flow: no ticks.')
         return
 
     # Log the drink.  If the username is empty or invalid, the backend will
     # assign it to the default (anonymous) user.  The backend will assign the
     # drink to a keg.
-    d = self._backend.RecordDrink(tap_name, ticks=ticks, username=username,
-        pour_time=pour_time, duration=duration, auth_token=auth_token,
-        spilled=False)
+    try:
+      d = self._backend.RecordDrink(meter_name, ticks=ticks, username=username,
+          pour_time=pour_time, duration=duration, auth_token=auth_token,
+          spilled=False)
+    except backend.DoesNotExistException as e:
+      self._logger.info('No drink recorded: %s' % e)
+      return
 
     if not d:
       self._logger.warning('No drink recorded.')
@@ -420,7 +372,7 @@ class DrinkManager(Manager):
     created = kbevent.DrinkCreatedEvent()
     created.flow_id = flow_id
     created.drink_id = d.id
-    created.tap_name = tap_name
+    created.meter_name = meter_name
     created.start_time = d.time
     created.end_time = d.time
     created.username = username
@@ -493,17 +445,17 @@ class TokenRecord:
   STATUS_ACTIVE = 'active'
   STATUS_REMOVED = 'removed'
 
-  def __init__(self, auth_device, token_value, tap_name):
+  def __init__(self, auth_device, token_value, meter_name):
     self.auth_device = auth_device
     self.token_value = token_value
-    self.tap_name = tap_name
+    self.meter_name = meter_name
     self.status = self.STATUS_ACTIVE
 
   def __str__(self):
     return '%s=%s@%s' % self.AsTuple()
 
   def AsTuple(self):
-    return (self.auth_device, self.token_value, self.tap_name)
+    return (self.auth_device, self.token_value, self.meter_name)
 
   def SetStatus(self, status):
     self.status = status
@@ -534,7 +486,7 @@ class AuthenticationManager(Manager):
 
   @EventHandler(kbevent.TokenAuthEvent)
   def HandleAuthTokenEvent(self, event):
-    for tap in self._GetTapsForTapName(event.tap_name):
+    for tap in self._GetTapsForTapName(event.meter_name):
       record = self._GetRecord(event.auth_device_name, event.token_value,
           tap.GetName())
       if event.status == event.TokenState.ADDED:
@@ -542,9 +494,9 @@ class AuthenticationManager(Manager):
       else:
         self._TokenRemoved(record)
 
-  def _GetRecord(self, auth_device, token_value, tap_name):
-    new_rec = TokenRecord(auth_device, token_value, tap_name)
-    existing = self._tokens.get(tap_name)
+  def _GetRecord(self, auth_device, token_value, meter_name):
+    new_rec = TokenRecord(auth_device, token_value, meter_name)
+    existing = self._tokens.get(meter_name)
     if new_rec == existing:
       return existing
     return new_rec
@@ -554,7 +506,7 @@ class AuthenticationManager(Manager):
 
     This will either start or renew a flow on the FlowManager."""
     username = None
-    tap_name = record.tap_name
+    meter_name = record.meter_name
     try:
       token = self._backend.GetAuthToken(record.auth_device, record.token_value)
       username = token.get('username')
@@ -572,7 +524,7 @@ class AuthenticationManager(Manager):
     max_idle = common_defs.AUTH_DEVICE_MAX_IDLE_SECS.get(record.auth_device)
     if max_idle is None:
       max_idle = common_defs.AUTH_DEVICE_MAX_IDLE_SECS['default']
-    self._flow_manager.StartFlow(tap_name, username=username,
+    self._flow_manager.StartFlow(meter_name, username=username,
         max_idle_secs=max_idle)
 
   def _MaybeEndFlow(self, record):
@@ -585,7 +537,7 @@ class AuthenticationManager(Manager):
       is_captive = common_defs.AUTH_DEVICE_CAPTIVE['default']
     if is_captive:
       self._logger.debug('Captive auth device, ending flow immediately.')
-      self._flow_manager.StopFlow(record.tap_name)
+      self._flow_manager.StopFlow(record.meter_name)
     else:
       self._logger.debug('Non-captive auth device, not ending flow.')
 
@@ -593,7 +545,7 @@ class AuthenticationManager(Manager):
   def _TokenAdded(self, record):
     """Processes a record when a token is added."""
     self._logger.info('Token attached: %s' % record)
-    existing = self._tokens.get(record.tap_name)
+    existing = self._tokens.get(record.meter_name)
 
     if existing == record:
       # Token is already known; nothing to do except update it.
@@ -604,40 +556,25 @@ class AuthenticationManager(Manager):
       self._logger.info('Removing previous token')
       self._TokenRemoved(existing)
 
-    self._tokens[record.tap_name] = record
+    self._tokens[record.meter_name] = record
     self._MaybeStartFlow(record)
 
   @util.synchronized
   def _TokenRemoved(self, record):
     self._logger.info('Token detached: %s' % record)
-    if record != self._tokens.get(record.tap_name):
+    if record != self._tokens.get(record.meter_name):
       self._logger.warning('Token has already been removed')
       return
 
     record.SetStatus(record.STATUS_REMOVED)
-    del self._tokens[record.tap_name]
+    del self._tokens[record.meter_name]
     self._MaybeEndFlow(record)
 
-  def _GetTapsForTapName(self, tap_name):
-    if tap_name == common_defs.ALIAS_ALL_TAPS:
+  def _GetTapsForTapName(self, meter_name):
+    if meter_name == common_defs.ALIAS_ALL_TAPS:
       return self._tap_manager.GetAllTaps()
     else:
-      if self._tap_manager.TapExists(tap_name):
-        return [self._tap_manager.GetTap(tap_name)]
-      else:
-        return []
-
-
-class SubscriptionManager(Manager):
-  def __init__(self, event_hub):
-    super(SubscriptionManager, self).__init__(event_hub)
-    self._client = kegnet.KegnetClient()
-
-  @EventHandler(kbevent.DrinkCreatedEvent)
-  @EventHandler(kbevent.FlowUpdate)
-  @EventHandler(kbevent.SetRelayOutputEvent)
-  def RepostEvent(self, event):
-    if isinstance(event, kbevent.FlowUpdate):
-      if event.state == event.FlowState.COMPLETED:
-        return
-    self._client.SendMessage(event)
+      tap = self._tap_manager.GetTap(meter_name)
+      if tap:
+        return [tap]
+      return None
